/*

Title:		rc_mosaic.c
Author:		Shawn A. Clifford 	<sac@ufl.edu>
Date:		22-Jun-1994
Purpose:	Provide a remote control interface for a mosaic client.  We
		accomplish this by:
		
		1) invoke this program, which in turn will ...
		2) fork two processes to run Mosaic on specified machines;
		these processes will have the names 'RC_DISPLAY' and
		'RC_MANAGER'
		3) with the pid of the child process, create the
		/tmp/Mosaic.pid file used for remote control.  For VMS, this
		is done with a mailbox named 'MOSAIC_<username>'
		4) write 'goto' or 'newwin' commands with their following
		URLs to the temp file or mailbox
		5) send SIGUSR1 signal to the child, causing it to interrupt
		what it's doing and process the temp file (Un*x only)
		6) look for the "death of child" signal, SIGCHLD, upon which
		we will delete the /tmp/Mosaic.pid file and exit gracefully

Usage:		rc_mosaic [ RC_MANAGER_URL [ RC_DISPLAY_URL ]]

Note:		Steps 3,4,5 are being carried out by the RC_MOSAIC-CGI CGI
		script, for now.


Modification History:

When		Who	Description
------------------------------------------------------------------------------
22-Jun-1994	SAC	First attempt
28-Feb-1995	SAC	Added socket support to interface with CGI

*/



/* Defines */

				/* LOCAL:  Where is your Mosaic binary? */
#ifndef VMS
#	define	MOSAIC_CMD "/local/X11/bin/Mosaic/Mosaic-sun\0"
#else
#	define	MOSAIC_CMD "public$:[x11.mosaic]mosaic\0"
#endif

				/* LOCAL:  Arguments for Mosaic (homepage) */
#define MOSAIC_ARG "http://alpha.nuceng.ufl.edu/\0"


				/* LOCAL:  Prefix for Mosaic control file */
#ifndef VMS
#	define MOSAIC_CTL "/tmp/Mosaic."
#else
#	define MOSAIC_CTL "MOSAIC_" /* Mailbox name; append username later */
#endif

#define DEBUG 0			/* Turn on/off some simple feedback */



/* Prototypes */

void sigchld_handler();		/* Death-of-child/exit handler */
void cleanup(int, int, int);	/* Sockets cleanup */
void check_display();		/* Check environment variables */



/* Fixup signals for VMS, SGI, and other System V platforms */

#ifndef  SIGCHLD		/* System V (SGI) */
#   if defined(SIGCLD)
#	define SIGCHLD SIGCLD	/* SIGCHLD is the BSD signal name */
#   else
#	define SIGCHLD 20	/* Death of child is not defined for VMS */
#   endif
#endif

#ifdef sgi
#	define _BSD_SIGNALS	/* Let's _please_ use BSD signals on SGI */
#endif




/* Headers */

#include "rc_mosaic.h"		/* LOCAL: Common defines for RC_MOSAIC */
				/* Edit this file to change the default port */

#include <stdio.h>		/* Normal stuff */
#include <stdlib.h>		/* exit(), etc. */
#include <string.h>		/* memset(), strlen(), etc. */
#include <errno.h>		/* Error codes & perror(), etc. */
#include <signal.h>		/* Signal defs. */

#ifdef VMS
#	include "vmsdefs.h"	/* VMS defines/includes/etc. */
#else
#	include <sys/types.h>
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <netdb.h>
#	include <arpa/inet.h>
#	include <sys/uio.h>
#	include <unistd.h>	/* unlink(), access(), ... */
	static	char
	   val_buffer[80];	/* Misc. string; passes info to CGI */
	static	char
	   *envp[2] = {
	   (char *)&val_buffer,	/* String for DISPLAY environment variable */
	   (char *)NULL		/* End of environment marker */
	   };
#endif

#ifdef sparc
#	include <vfork.h>	/* Fixup vfork() stuff during global opt. */
#endif




/* Global vars. */

int sigflag = FALSE;		/* Global variable set when signal occurs */
int childpid;	        	/* Pid of the RC_DISPLAY process */
int mgr_pid;			/* Pid of the RC_MANAGER process */
static int sock_1, sock_2;	/* Sockets */



int main(int argc, char **argv)
{
static 	struct  sockaddr_in sock1_name;         /* Address struct for sock_1.*/
static 	struct  sockaddr_in retsock1_name;      /* Address struct for sock_1.*/
        struct  hostent         hostentstruct;  /* Storage for hostent data.  */
        struct  hostent         *hostentptr;    /* Pointer to hostent data.   */
        static  char            hostname[256];  /* Name of local host.        */
	int	retval;				/* helpful for debugging */
	int	flag;
	int	namelength;

   /* Step 1:  Invoke this program (done!) */
   /* ==================================== */

   check_display();		/* Check RC_xxxxxxx environment variables */
   if (DEBUG) printf("\n%s\n\n", ident);


   
   /* Step 2:  Startup the child process */
   /* ================================== */
   


   /* Mark the "death of child signal" for trapping */

#ifndef VMS
   signal(SIGCHLD, sigchld_handler);
#else
   declare_exit_handler();	/* In case of an unusual event; like d.o.c. */
#endif
   


   /* Fork off child processes. */


   /* RC_MANAGER CHILD PROCESS */  
   /* ------------------------ */

      
#ifndef VMS
      
   if ( (mgr_pid = vfork()) == -1) {
      perror("can't vfork RC_MANAGER");
      _exit(1);
   } else if (mgr_pid == 0) {


      /* Disassociate from control terminal */
      /* This will allow us to put this program in the background. */

      setpgrp();		/* change process group and lose ctrl tty */
      
      if (DEBUG) {
	 printf("child (RC_Manager):  child pid = %d, parent pid = %d\n",
		childpid=getpid(), getppid());	
	 if (argc > 1) printf("Arg[1]: %s\n", argv[1]);
      }
	      

      /* Set up the environment string for DISPLAY */
	   
      memset(val_buffer, 0, sizeof(val_buffer));
      sprintf(val_buffer, "DISPLAY=%s", getenv("RC_MANAGER"));

      /* Kick off the Presentation Manager Mosaic to RC_MANAGER */

      if (argc > 1) {
	 if ( (execle(MOSAIC_CMD, "RC_Manager\0", argv[1],
			(char *)0, envp)) == -1) {
			perror("**Can not exec");
			_exit(1);
	 }
      }
      else
	 if ( (execle(MOSAIC_CMD, "RC_Manager\0", MOSAIC_ARG,
			(char *)0, envp)) == -1) {
			perror("**Can not exec");
			_exit(1);
	      
	 }

   };	/* RC_MANAGER process */

#else	/* VMS: We could use execl, but it is less robust on VMS than spawn */

      /* Temporarily redefine DECW$DISPLAY to RC_MANAGER */
	
      lib$set_logical(&display, &rc_manager, &table);

      /* Spawn the presentation manager Mosaic, RC_MANAGER*/

      if (argc > 1)		/* If there was a URL passed, then use it. */
	 sprintf(spawn_buff, "MCR %s %s\0", MOSAIC_CMD, argv[1]);
      else
	 sprintf(spawn_buff, "MCR %s %s\0", MOSAIC_CMD, MOSAIC_ARG);
      spawn_cmd.dsc$w_length = strlen(spawn_buff);
      retstat = lib$spawn(&spawn_cmd, &spawn_nul, &spawn_nul, &spawn_flag,
		&rc_manager, &mgr_pid, 0,0, &sigchld_handler, 0,0,0);

      /* Restore the display to DECW$DISPLAY */
	
      lib$delete_logical(&display, &table);

      if (DEBUG)
	 printf("Spawn command: %s\n", spawn_buff);

#endif




   /* RC_DISPLAY CHILD PROCESS */
   /* ------------------------ */


#ifndef VMS	

   if ( (childpid = vfork()) == -1) {
      perror("can't vfork RC_DISPLAY");	
      _exit(1);
   } else if (childpid == 0) {
	    
      setpgrp();		/* change process group and lose ctrl tty */
	     
      if (DEBUG) {
	 printf("child (RC_Display):  child pid = %d, parent pid = %d\n",
		childpid=getpid(), getppid());
	 if (argc > 2) printf("Arg[2]: %s\n", argv[2]);
      }

      /* Set up the environment string for DISPLAY */

      memset(val_buffer, 0, sizeof(val_buffer));
      sprintf(val_buffer, "DISPLAY=%s", getenv("RC_DISPLAY"));

      /* Kick off the slave Mosaic to RC_DISPLAY */
	
      if (argc > 2) {
	   if ((execle(MOSAIC_CMD, "RC_Display\0", argv[2],
		(char *)0, envp)) == -1) {
		perror("**Can not exec");
		_exit(1);	/* Exit RC_DISPLAY.  Use _exit, NOT exit */
	   }
      } else if ((execle(MOSAIC_CMD, "RC_Display\0", MOSAIC_ARG,
		(char *)0, envp)) == -1) {
		perror("**Can not exec");
		_exit(1);	/* Exit RC_DISPLAY.  Use _exit, NOT exit */
      }


   } else {

#else	/* VMS: We could use execl, but it is less robust on VMS than spawn */
	
      /* Temporarily redefine DECW$DISPLAY to RC_DISPLAY */
	
      lib$set_logical(&display, &rc_display, &table);

      /* Spawn the slave Mosaic, using SIGCHLD as the event flag */

      if (argc > 2)		/* If there was a URL passed, then use it.*/
	 sprintf(spawn_buff, "MCR %s -mbx %s\0", MOSAIC_CMD, argv[2]);
      else
	 sprintf(spawn_buff, "MCR %s -mbx %s\0", MOSAIC_CMD, MOSAIC_ARG);
      spawn_cmd.dsc$w_length = strlen(spawn_buff);
      retstat = lib$spawn(&spawn_cmd, &spawn_nul, &spawn_nul, &spawn_flag,
		&rc_display, &childpid, 0, &spawn_ef, &sigchld_handler, 0,0,0);

      /* Restore the display to DECW$DISPLAY */
	
      lib$delete_logical(&display, &table);

      if (DEBUG) {
	 printf("Spawn command: %s\n", spawn_buff);
	 printf("child (RC_Manager):  pid = %d (0x%X)\n", mgr_pid, mgr_pid);
	 printf("child (RC_Display):  pid = %d (0x%X)\n", childpid, childpid);
      }
   
#endif
	
      /* PARENT PROCESS */
      /* -------------- */


      /*
       * Open socket 1: AF_INET, SOCK_STREAM. 
       */
      if ((sock_1 = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
	 perror( "socket");
	 exit(1);
      }

      /*
       * Get the host local name. 
       */
      retval = gethostname(hostname,sizeof hostname);
      if (retval) {
	 perror ("gethostname");
	 cleanup (1, sock_1, 0);
      }

      /*
       * Get pointer to network data structure for socket 1. 
       */
      if ((hostentptr = gethostbyname (hostname)) == NULL) {
	 perror( "gethostbyname");
	 cleanup(1, sock_1, 0);
      }

      /*
       * Copy hostent data to safe storage. 
       */
      hostentstruct = *hostentptr;

      /*
       * Fill in the name & address structure for socket 2. 
       */
      sock1_name.sin_family = hostentstruct.h_addrtype;
      sock1_name.sin_port = htons(RC_MOSAIC_PORT);
      sock1_name.sin_addr = * ((struct in_addr *) hostentstruct.h_addr);

      /*
       * Bind name to socket 1.
       */
      retval = bind (sock_1, &sock1_name, sizeof sock1_name);
      if (retval) {
	 perror("bind");
	 cleanup(1, sock_1, 0);
      }


      /*
       * Listen on socket 1 for connections. 
       */
      retval = listen (sock_1, 5);
      if (retval) {
	 perror("listen");
	 cleanup(1, sock_1, 0);
      }


      /* Create the control file name or (for VMS) get the mailbox name */

#ifdef VMS

      memset(username, 0, sizeof(username));
      memset(log_buffer, 0, sizeof(log_buffer));
      memset(val_buffer, 0, sizeof(val_buffer));
      sys$getjpiw(0,0,0,&itmlst,0,0,0);	/* Get our username */
      for (i=sizeof(username)-1; i>0; i--)	/* Convert blanks to nulls */
	 if (username[i]==' ') username[i]='\0';
      sprintf(log_buffer, "%s%s", MOSAIC_CTL, username);
      logical.dsc$w_length = strlen(log_buffer);
      while(val_buffer[0] == '\0')		/* Translate MBX_<username> */
	 sys$trnlnm(&attr, &table, &logical, 0, &lnm_itmlst);
      if (DEBUG)
	 printf("Control mailbox:  %s = %s\n", log_buffer, val_buffer);

      /* Assign a channel to the mailbox */
      retstat = sys$assign(&mbox, &chan, 0, 0, 0);
      if (!(retstat & 0x00000001))
	 lib$signal(retstat);

      /* Change the protection on the mailbox so everyone can write to it */
      func = IO$_SETMODE + IO$M_SETPROT;
      retstat = sys$qiow(0, chan, func, 0, 0, 0, 0, prot_mask, 0, 0, 0, 0);
      if (!(retstat & 0x00000001))
	 lib$signal(retstat);

      /* Deassign the channel */
      retstat = sys$dassgn(chan);

#else

      /* Write the pid to the file name to send to the CGI-script client */
      sprintf(val_buffer, "%s%d\0", MOSAIC_CTL, childpid);

#endif


      /* This is the server section. */

      while (sigflag == FALSE) {

	   /*
	    * Accept connection from socket 1:
	    * accepted connection will be on socket 2. 
	    */
	   namelength = sizeof (sock1_name);
	   sock_2 = accept(sock_1, (struct sockaddr *) &sock1_name,
		&namelength);


	   /* Send the pid or mailbox name to the client CGI script */

	   flag = 0;		/* maybe 0 or MSG_OOB */
	   retval = send(sock_2, val_buffer, strlen (val_buffer), flag);
	   if (retval < 0) {
		perror ("send()");
		exit(1);
	   }

	   close(sock_2);

      } /* while() */




#if 0	/* This section is #if'd out until it can work asynchronously */

	
	/* Wait for the "death of child" signal */
	if (DEBUG)
	   printf("parent (RC_MOSAIC): waiting for the child to die ...\n");

#ifdef VMS

	sys$waitfr(SIGCHLD);	/* Wait for this event flag to occur */

#elif (defined(SIGTSTP) || defined(__DECC)) /* BSD or VMS's DECC (weird) */

/*	sigblock(sigmask(SIGCHLD));	/* (BSD) is equivalent to... */
	sigsetmask(SIGCHLD);	/* which is supported on VMS    */
	while (sigflag == FALSE)/* but SIGCHLD isn't, so use event flags */
	   sigpause(0);		/* Wait for a signal to occur */
	     
	     /* ... signal happened, process it ... */
	     
#else				/* System V */
	
	sighold(SIGCHLD);	
	while (sigflag == FALSE)
	   sigpause(SIGCHLD);	/* Wait for a signal to occur */
	sigrelse(SIGCHLD);
	     
	     /* ... signal happened, process it ... */
	
#endif
	
#endif  /* #if 0 */


#ifndef VMS
   };	/* if vfork() ... */
#endif



   /* Cleanup code; just in case we didn't get run down by sigchld_handler */
   cleanup(2, sock_1, sock_2);
   return 1;			/* Th-th-that's all folks! */

}	/* End of main() */



/* -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+- */



/*******************
   sigchld_handler
 *******************/

void sigchld_handler()
{
char ctl_file[80];
   
   if (!(sigflag)) {
     
      sigflag = TRUE;		/* Mark the signal */

      if (DEBUG) {
         printf("parent (RC_MOSAIC):  child process has exited...\
 cleaning up and exiting\n");
	 printf("Childpid: %d (0x%X), Mgr Pid: %d (0x%X)\n",childpid,childpid,
		mgr_pid, mgr_pid);
      }

      /* Stop the RC_MANAGER/RC_DISPLAY processes */
      kill(mgr_pid, SIGHUP);
      kill(childpid, SIGHUP);

      /* Cleanup the sockets */
      cleanup(2, sock_1, sock_2);
   }


#ifndef VMS


   /* Step 6:  Delete control file on death-of-child */
   /* ============================================== */


   /* Delete /tmp/Mosaic.<pid> */
   
   sprintf (ctl_file, "%s%d\0", MOSAIC_CTL, childpid);

   
   /* Does the file exist? */

   if (access(ctl_file, F_OK) == 0)
     {
	if (DEBUG) printf ("Deleting %s\n", ctl_file);
   	unlink (ctl_file);	/* use delete() on VMS */
     };

#else

/*   gsignal(SIGCHLD);		/* Release the sigpause(0), if */
				/* you ever get it to work */
#endif



   /* We don't want to return to the main code, else we will get */
   /* erroneous errors from the accept() or send() calls. */
   /* We use _exit(), instead of exit(), because it is correct for use */
   /* with vfork()/fork() processes */

   _exit(1);   

}	/* sigchld_handler() */




/***********
   cleanup
 ***********/

void cleanup(int how_many, int sock1, int sock2)
{
	int	retval;

	/*
	 * Shutdown and close sock1 completely.
	 */
	retval = shutdown(sock1,2);
        retval = close (sock1);


	/*
	 * If given, shutdown and close sock2. 
	 */
	if (how_many == 2)
	{
		retval = shutdown(sock2,2);
        	retval = close (sock2);
	}

}	/* end cleanup*/




/*****************
   check_display
 *****************/

void check_display()
{

   if ( (getenv("RC_DISPLAY") == NULL) || (getenv("RC_MANAGER") == NULL) ) {
      printf("\nYou must define the environment variables RC_MANAGER and ");
      printf("RC_DISPLAY before\nrunning this program.\n\n");
      printf("RC_MANAGER should point to the X display that will be ");
      printf("used as the presentation\n   manager Mosaic.\n\n");
      printf("RC_DISPLAY should point to the X display that will be used as");
      printf(" the client,\n   or display, Mosaic.\n\n");
      printf("To define these variables, use the following format:\n\n");

#ifndef VMS
      printf("\tsetenv RC_MANAGER node1.domain.site.net:0\n");
      printf("\tsetenv RC_DISPLAY node2.domain.site.net:0\n");
      printf("or\n");
      printf("\tsetenv RC_MANAGER $DISPLAY\n");
      printf("\t(if you want RC_MANAGER to be your current display)\n\n");
#else
      printf("   set display/create/node=node1.domain.site.net/");
      printf("transport=tcpip RC_MANAGER\n");
      printf("   set display/create/node=node2.domain.site.net/");
      printf("transport=tcpip RC_DISPLAY\n");
      printf("or\n");
      printf("   define RC_MANAGER 'f$trnlnm(\"DECW$DISPLAY\")\n");
      printf("   (if you want RC_MANAGER to be your current display)\n\n");
#endif

      printf("where 'node1.domain.site.net' is an Internet address.\n\n");

      exit(1);
   }

   return;
}
